/*  Copyright 1997 Regents of the University of California.
Permission is granted to use this software for any noncommercial purpose.
For commercial licensing contact the UCSD Technology Transfer Office./

UC MAKES NO WARRANTY, ESPRESS OR IMPLIED, IN CONNECTIONS WITH THIS SOFTWARE!

Written by Miller Puckette, msp@ucsd.edu +1 619 534 4823 */


/* 
 * Fiddle is a pitch tracker hardwired to have hop size ("H") equal to
 * half its window size ("N").
 * 
 * The "lastanalysis" field holds the shifted FT of the previous H
 * samples.  The buffer contains in effect points 1/2,  3/2, ..., (H-1)/2
 * of the FT of a real vector of length N, half of whose points are zero, 
 * i.e.,  only the first H points are used.  Put another way, we get the
 * the odd-numbered points of the FFT of the H points, zero padded to 4*H in
 * length. The "real" points 0, 1, ..., H-1
 * are found by interpolating these others,  using the fact that the
 * half-integer points are band-limited (they only have positive frequencies.)
 * To facilitate the interpolation the "lastanalysis" buffer contains
 * FILTSIZE extra points (1/2-FILTSIZE, ...,  -1/2) at the beginning and
 * FILTSIZE again at the end ((H+1)/2, ..., FILTSIZE+(H-1)/2).  The buffer
 * is therefore of size N+4*FILTSIZE.
 *
 * this version should compile for Max, FTS, or PD.


 * DOLIST
 * new-note attacks
 * peak bifurcation
 * should nhist really be npitch?
 */

/*#if defined( NT ) (__GNU_LIBRARY__ == 1) */
/* NT and old glib on linux doesn't appear 
to have single-precision ANSI math */

#define flog log
#define fexp exp
#define fsqrt sqrt
#pragma warning (disable: 4305 4244)
/*#endif*/

char fiddle_version[] = "fiddle version 0.7";

#include "m_pd.h"

#include <math.h>


#define MINBIN 3
#define DEFAMPLO 40
#define DEFAMPHI 50
#define DEFATTACKTIME 100
#define DEFATTACKTHRESH 10
#define DEFVIBTIME 50
#define DEFVIBDEPTH 0.5
#define GLISS 0.7f
#define DBFUDGE 30.8f

#define MAXNPITCH 3
#define MAXHIST 3		    /* find N hottest peaks in histogram */

#define MAXPOINTS 1024
#define MINPOINTS 128
#define DEFAULTPOINTS 1024

#define HISTORY 20
#define MAXLOWPEAK 20		/* pick off NLOW of the lowest-freq peaks */
#define MAXSTRONGPEAK 0		/* ... and then the NSTRONG strongest left */
#define MAXPEAK (MAXLOWPEAK + MAXSTRONGPEAK)
#define MINBW (0.03f)			/* consider BW >= 0.03 FFT bins */

#define BINPEROCT 48			/* bins per octave */
#define BPERO_OVER_LOG2 69.24936196f	/* BINSPEROCT/log(2) */
#define FACTORTOBINS (float)(4/0.0145453)	/* 4 / (pow(2.,1/48.) - 1) */
#define BINGUARD 10			/* extra bins to throw in front */
#define PARTIALDEVIANCE 0.023f		/* acceptable partial detuning in % */
#define LOGTODB 4.34294481903f		/* 20/log(10) */

#define KNOCKTHRESH 10.f     /* don't know how to describe this */


static float sigfiddle_partialonset[] =
{
0, 
48,
76.0782000346154967102, 
96,
111.45254855459339269887, 
124.07820003461549671089, 
134.75303625876499715823, 
144,
152.15640006923099342109, 
159.45254855459339269887, 
166.05271769459026829915, 
172.07820003461549671088, 
177.62110647077242370064, 
182.75303625876499715892, 
187.53074858920888940907, 
192,
};

#define NPARTIALONSET (sizeof(sigfiddle_partialonset)/sizeof(float))

static int sigfiddle_intpartialonset[] =
{
0, 
48,
76, 
96,
111, 
124, 
135, 
144,
152, 
159, 
166, 
172, 
178, 
183, 
188, 
192,
};


#if 0
static char sigfiddle_hmask[] =
{
    1, 1, 0, 0, 0, 0, 0, 0, 0, 0,   /* 0 */
    0, 0, 0, 0, 0, 0, 0, 0, 0, 0,   /* 10 */
    0, 0, 0, 0, 0, 0, 0, 0, 0, 0,   /* 20 */
    0, 0, 0, 0, 0, 0, 0, 0, 0, 0,   /* 30 */
    0, 0, 0, 0, 0, 0, 0, 1, 1, 1,   /* 40 */
    0, 0, 0, 0, 0, 0, 0, 0, 0, 0,   /* 50 */
    0, 0, 0, 0, 0, 0, 0, 0, 0, 0,   /* 60 */
    0, 0, 0, 0, 0, 1, 1, 1, 0, 0,   /* 70 */
    0, 0, 0, 0, 0, 0, 0, 0, 0, 0,   /* 80 */
    0, 0, 0, 0, 0, 1, 1, 1, 0, 0,   /* 90 */
    0, 0, 0, 0, 0, 0, 0, 0, 0, 0,   /* 100 */
    1, 1, 1, 0, 0, 0, 0, 0, 0, 0,   /* 110 */
    0, 0, 0, 1, 1, 1, 0, 0, 0, 0,   /* 120 */
    0, 0, 0, 0, 1, 1, 1, 0, 0, 0,   /* 130 */
    0, 0, 0, 1, 1, 1, 0, 0, 0, 0,   /* 140 */
    0, 0, 1, 0, 0, 0, 0, 0, 0, 1,   /* 150 */
    0, 0, 0, 0, 0, 0, 1, 0, 0, 0,   /* 160 */
    0, 0, 1, 0, 0, 0, 0, 0, 1, 0,   /* 170 */
    0, 0, 1, 0, 0, 0, 0, 0, 1, 0,   /* 180 */
    0, 0, 1, 0, 0, 0, 0, 0, 0, 0,   /* 190 */
};
#endif

#define HMASKSIZE 200

#define MAXPOSSIBLEBIN 480
static float sigfiddle_binbin[MAXPOSSIBLEBIN];


/* these come from the "upsamp" subdirectory: */
#define FILT1 ((float)(.5 * 1.227054))
#define FILT2 ((float)(.5 * -0.302385))
#define FILT3 ((float)(.5 * 0.095326))
#define FILT4 ((float)(.5 * -0.022748))
#define FILT5 ((float)(.5 * 0.002533))
#define FILTSIZE 5


typedef struct peak
{
    float p_freq;		    /* frequency in bins */
    float p_width;		    /* peak width in bins */
    float p_pow;		    /* peak power */
    float *p_fp;		    /* pointer back to spectrum */
} t_peak;

typedef struct histopeak
{
    float h_pitch;		    /* estimated pitch */
    float h_value;		    /* value of peak */
    float h_loud;		    /* combined strength of found partials */
    int h_index;		    /* index of bin holding peak */
    int h_used;			    /* true if an x_hist entry points here */
} t_histopeak;

typedef struct pitchhist	    /* struct for keeping history by pitch */
{
    float h_pitch;		    /* pitch to output */
    float h_amps[HISTORY];	    /* past amplitudes */
    float h_pitches[HISTORY];	    /* past pitches */
    float h_noted;		    /* last pitch output */
    int h_age;			    /* number of frames pitch has been there */
    t_histopeak *h_wherefrom;	    /* new histogram peak to incorporate */
    void *h_outlet;
} t_pitchhist;

typedef struct sigfiddle		    /* instance struct */
{
    t_object x_ob;		    /* object header */
    t_clock *x_clock;    	    /* callback for timeouts */

    float *x_inbuf;		    /* buffer to analyze, npoints/2 elems */
    float *x_lastanalysis;	    /* FT of last buffer (see main comment) */
    float *x_spiral;		    /* 1/4-wave complex exponential */
    int x_phase;		    /* number of points since last output */
    int x_histphase;		    /* phase into amplitude history vector */
    int x_hop;			    /* period of output, npoints/2 */
    float x_sr;			    /* sample rate */
    t_pitchhist x_hist[MAXNPITCH];  /* history of current pitches */
    int x_nprint;		    /* how many periods to print */
    int x_npitch;		    /* number of simultaneous pitches */
    int x_nhist;		    /* number of histo-peaks */
    float x_dbs[HISTORY];	    /* DB history, indexed by "histphase" */
    float x_peaked;		    /* peak since last attack */
    int x_dbage;		    /* number of bins DB has met threshold */
/* parameters */
    float x_amplo;
    float x_amphi;
    float x_attacktime;
    float x_attackbins;
    float x_attackthresh;
    float x_vibtime;
    float x_vibbins;
    float x_vibdepth;
    float x_npartial;
/* outlets & clock */
    void *x_envout;
    float x_attackvalue;
    void *x_attackout;
    void *x_noteout;
} t_sigfiddle;

#if CHECKER
float fiddle_fucker[1024];
#endif

int sigfiddle_ilog2(int n)
{
    int ret = -1;
    while (n)
    {
		n >>= 1;
		ret++;
    }
    return (ret);
}

float fiddle_mtof(float f)
{
	return (8.17579891564 * exp(.0577622650 * f));
}

float fiddle_ftom(float f)
{
	return (17.3123405046 * log(.12231220585 * f));
}
#define ftom fiddle_ftom
#define mtof fiddle_mtof

void sigfiddle_doit(t_sigfiddle *x)
{
    float spect1[4*MAXPOINTS];
    float spect2[MAXPOINTS + 4*FILTSIZE];
#if CHECKER
    float fucker3[4*MAXPOINTS];
#endif

    t_peak peaklist[MAXPEAK + 1], *pk1, *hipeak = peaklist+MAXLOWPEAK;
    t_histopeak histvec[MAXHIST], *hp1;
    int i, j, k,  hop = x->x_hop, n = 2*hop, npeak, nhist,
        logn = sigfiddle_ilog2(n), newphase, oldphase;
    float *fp, *fp1, *fp2, *fp3, pow, db;
    float maxbin = BINPEROCT * (logn-2),  *histogram = spect2 + BINGUARD;
    t_pitchhist *phist;
    float hzperbin = x->x_sr / (2.0f * n);

    oldphase = x->x_histphase;
    newphase = x->x_histphase + 1;
    if (newphase == HISTORY) newphase = 0;
    x->x_histphase = newphase;

	/*
	 * multiply the H points by a 1/4-wave complex exponential,
	 * and take FFT of the result.
	 */
    for (i = 0, fp1 = x->x_inbuf, fp2 = x->x_spiral, fp3 = spect1;
	i < hop; i++, fp1++, fp2 += 2, fp3 += 2)
	    fp3[0] = fp1[0] * fp2[0], fp3[1] = fp1[0] * fp2[1];

    pd_fft(spect1, hop, 0);

	/*
	 * now redistribute the points to get in effect the odd-numbered
	 * points of the FFT of the H points, zero padded to 4*H in length.
	 */
    for (i = 0, fp1 = spect1, fp2 = spect2 + (2*FILTSIZE);
	i < (hop>>1); i++, fp1 += 2, fp2 += 4)
	    fp2[0] = fp1[0], fp2[1] = fp1[1];
    for (i = 0, fp1 = spect1 + n - 2, fp2 = spect2 + (2*FILTSIZE+2);
	i < (hop>>1); i++, fp1 -= 2, fp2 += 4)
	    fp2[0] = fp1[0], fp2[1] = -fp1[1];
    for (i = 0, fp1 = spect2 + (2*FILTSIZE), fp2 = spect2 + (2*FILTSIZE-2);
	i<FILTSIZE; i++, fp1+=2, fp2-=2)
	    fp2[0] = fp1[0],  fp2[1] = -fp1[1];
    for (i = 0, fp1 = spect2 + (2*FILTSIZE+n-2), fp2 = spect2 + (2*FILTSIZE+n);
	i<FILTSIZE; i++, fp1-=2, fp2+=2)
	    fp2[0] = fp1[0],  fp2[1] = -fp1[1];
#if 0
    {
	fp = spect2 + 2*FILTSIZE;
	post("x1 re %12.4f %12.4f %12.4f %12.4f %12.4f",
	    fp[0], fp[2], fp[4], fp[6], fp[8]);
	post("x1 im %12.4f %12.4f %12.4f %12.4f %12.4f",
	    fp[1], fp[3], fp[5], fp[7], fp[9]);
    }
#endif	    
	/* spect2 is now prepared; now combine spect2 and lastanalysis into
	 * spect1.  Odd-numbered points of spect1 are the points of "last"
	 * plus (-i, i, -i, ...) times spect1.  Even-numbered points are
	 * the interpolated points of "last" plus (1, -1, 1, ...) times the
	 * interpolated points of spect1.
	 *
	 * To interpolate, take FILT1 exp(-pi/4) times
	 * the previous point,  FILT2*exp(-3*pi/4) times 3 bins before,
	 * etc,  and FILT1 exp(pi/4), FILT2 exp(3pi/4), etc., to weight
	 * the +1, +3, etc., points.
	 *
	 * In this calculation,  we take (1, i, -1, -i, 1) times the
	 * -9, -7, ..., -1 points, and (i, -1, -i, 1, i) times the 1, 3,..., 9
	 * points of the OLD spectrum, alternately adding and subtracting
	 * the new spectrum to the old; then we multiply the whole thing
	 * by exp(-i pi/4).
	 */
    for (i = 0, fp1 = spect1, fp2 = x->x_lastanalysis + 2*FILTSIZE,
	fp3 = spect2 + 2*FILTSIZE;
	    i < (hop>>1); i++)
    {
	float re,  im;

	re= FILT1 * ( fp2[ -2] -fp2[ 1]  +fp3[ -2] -fp3[ 1]) +
	    FILT2 * ( fp2[ -3] -fp2[ 2]  +fp3[ -3] -fp3[ 2]) +
	    FILT3 * (-fp2[ -6] +fp2[ 5]  -fp3[ -6] +fp3[ 5]) +
	    FILT4 * (-fp2[ -7] +fp2[ 6]  -fp3[ -7] +fp3[ 6]) + 
	    FILT5 * ( fp2[-10] -fp2[ 9]  +fp3[-10] -fp3[ 9]);

	im= FILT1 * ( fp2[ -1] +fp2[ 0]  +fp3[ -1] +fp3[ 0]) +
	    FILT2 * (-fp2[ -4] -fp2[ 3]  -fp3[ -4] -fp3[ 3]) +
	    FILT3 * (-fp2[ -5] -fp2[ 4]  -fp3[ -5] -fp3[ 4]) +
	    FILT4 * ( fp2[ -8] +fp2[ 7]  +fp3[ -8] +fp3[ 7]) + 
	    FILT5 * ( fp2[ -9] +fp2[ 8]  +fp3[ -9] +fp3[ 8]);

	fp1[0] = 0.7071f * (re + im);
	fp1[1] = 0.7071f * (im - re);
	fp1[4] = fp2[0] + fp3[1];
	fp1[5] = fp2[1] - fp3[0];
	
	fp1 += 8, fp2 += 2, fp3 += 2;
	re= FILT1 * ( fp2[ -2] -fp2[ 1]  -fp3[ -2] +fp3[ 1]) +
	    FILT2 * ( fp2[ -3] -fp2[ 2]  -fp3[ -3] +fp3[ 2]) +
	    FILT3 * (-fp2[ -6] +fp2[ 5]  +fp3[ -6] -fp3[ 5]) +
	    FILT4 * (-fp2[ -7] +fp2[ 6]  +fp3[ -7] -fp3[ 6]) + 
	    FILT5 * ( fp2[-10] -fp2[ 9]  -fp3[-10] +fp3[ 9]);

	im= FILT1 * ( fp2[ -1] +fp2[ 0]  -fp3[ -1] -fp3[ 0]) +
	    FILT2 * (-fp2[ -4] -fp2[ 3]  +fp3[ -4] +fp3[ 3]) +
	    FILT3 * (-fp2[ -5] -fp2[ 4]  +fp3[ -5] +fp3[ 4]) +
	    FILT4 * ( fp2[ -8] +fp2[ 7]  -fp3[ -8] -fp3[ 7]) + 
	    FILT5 * ( fp2[ -9] +fp2[ 8]  -fp3[ -9] -fp3[ 8]);

	fp1[0] = 0.7071f * (re + im);
	fp1[1] = 0.7071f * (im - re);
	fp1[4] = fp2[0] - fp3[1];
	fp1[5] = fp2[1] + fp3[0];
	
	fp1 += 8, fp2 += 2, fp3 += 2;
    }
#if 0
    if (x->x_nprint)
    {
	for (i = 0,  fp = spect1; i < 16; i++,  fp+= 4)
	    post("spect %d %f %f --> %f", i, fp[0], fp[1],
		sqrt(fp[0] * fp[0] + fp[1] * fp[1]));
    }
#endif
	 /* copy new spectrum out */
    for (i = 0, fp1 = spect2, fp2 = x->x_lastanalysis;
	    i < n + 4*FILTSIZE; i++) *fp2++ = *fp1++;

    for (i = 0; i < MINBIN; i++) spect1[4*i + 2] = spect1[4*i + 3] = 0;
	/* starting at bin MINBIN, compute hanning windowed power spectrum */
    for (i = MINBIN, fp1 = spect1+4*MINBIN, pow = 0; i < n-2; i++,  fp1 += 4)
    {
	float re = fp1[0] - 0.5f * (fp1[-8] + fp1[8]);
	float im = fp1[1] - 0.5f * (fp1[-7] + fp1[9]);
	fp1[3] = (pow += (fp1[2] = re * re + im * im));
    }
    
    if (pow > 1e-9f)
    {
	db = (100.f - DBFUDGE) + LOGTODB * log(pow/n);
	if (db < 0) db = 0;
    }
    else db = 0;
	/*  store new db in history vector */
    x->x_dbs[newphase] = db;
    if (db < x->x_amplo) goto nopow;
#if 1
    if (x->x_nprint) post("power %f", pow);
#endif

#if CHECKER
    for (i = 0; i < hop; i++)
    {
	fucker3[2*i] = fiddle_fucker[i];
	fucker3[2*i + 1] = 0;
	fucker3[n + 2*i] = fiddle_fucker[i] = x->x_inbuf[i];
	fucker3[n + 2*i + 1] = 0;
    }
    for (i = 2*n; i < 4*n; i++) fucker3[i] = 0;
    fft(fucker3, 2*n, 0);
    if (x->x_nprint)
    {
	for (i = 0,  fp = fucker3; i < 16; i++,  fp += 2)
	    post("spect %d %f %f --> %f", i, fp[0], fp[1],
		sqrt(fp[0] * fp[0] + fp[1] * fp[1]));
    }

#endif
    npeak = 0;

	 /* search for peaks: first the low frequency ones. */
    for (i = MINBIN, fp = spect1+4*MINBIN, pk1 = peaklist;
	i < n-2 && npeak < MAXLOWPEAK; i++, fp += 4)
    {
	float height = fp[2], h1 = fp[-2], h2 = fp[6];
	float totalfreq, pfreq, f1, f2, m, var, stdev;
	
	if (height < h1 || height < h2 ||
	    h1 < 0.00001f*pow || h2 < 0.00001f*pow)
	    	continue;

	pfreq= ((fp[-8] - fp[8]) * (2.0f * fp[0] - fp[8] - fp[-8]) +
		(fp[-7] - fp[9]) * (2.0f * fp[1] - fp[9] - fp[-7])) /
		    (2.0f * height);
	f1=    ((fp[-12] - fp[4]) * (2.0f * fp[-4] - fp[4] - fp[-12]) +
		(fp[-11] - fp[5]) * (2.0f * fp[-3] - fp[5] - fp[-11])) /
		    (2.0f * h1) - 1;
	f2=    ((fp[-4] - fp[12]) * (2.0f * fp[4] - fp[12] - fp[-4]) +
		(fp[-3] - fp[13]) * (2.0f * fp[5] - fp[13] - fp[-3])) /
		    (2.0f * h2) + 1;

	m = 0.333333f * (pfreq + f1 + f2);
	var = 0.5f * ((pfreq-m)*(pfreq-m) + (f1-m)*(f1-m) + (f2-m)*(f2-m));

	totalfreq = i + m;
	if (var * pow > KNOCKTHRESH * height || var < 1e-30)
	{
#if 0
	    if (x->x_nprint)
		post("cancel: %.2f hz, index %.1f, pow %.5f, stdev=%.2f", 
		    totalfreq * hzperbin, BPERO_OVER_LOG2 * log(totalfreq) - 96, 
		     height, sqrt(var));
#endif
	    continue;
	}
	stdev = fsqrt(var);
	if (totalfreq < 4)
	{
	    if (x->x_nprint) post("oops: was %d,  freq %f, m %f, stdev %f h %f", 
		i,  totalfreq, m, stdev, height);
	    totalfreq = 4;
	}
	pk1->p_width = stdev;

	pk1->p_pow = height;
	pk1->p_fp = fp;
	pk1->p_freq = totalfreq;
	npeak++;
#if 1
	if (x->x_nprint)
	{
	    post("peak: %.2f hz. index %.1f, pow %.5f, stdev=%.2f", 
		pk1->p_freq * hzperbin,
		BPERO_OVER_LOG2 * log(pk1->p_freq) - 96, 
		 height, stdev);
	}
#endif
	pk1++;
    }


#ifdef SEARCH_BY_STRENGTH
	/* then the strongest remaining ones */
    if (npeak == MAXLOWPEAK)
	for (; i < n-2 && npeak < MAXPEAK; i++, fp += 4)
    {
	float height = fp[2], h1 = fp[-2], h2 = fp[6];
	pk1 = peaklist + (npeak-1);
	if (height < h1 || height < h2
	    || height < 0.0001f*pow || h1 < 0.0001f*pow || h2 < 0.0001f*pow)
		continue;
	if (npeak != MAXLOWPEAK && pk1->p_pow > height)
	    continue;
	while (pk1 >= hipeak && pk1->p_pow < height)
	{
	    pk1[1].p_pow = pk1[0].p_pow;
	    pk1[1].p_fp = pk1[0].p_fp;
	    pk1[1].p_freq = pk1[0].p_freq;
	    pk1--;
	}
	pk1[1].p_pow = height;
	pk1[1].p_fp = fp;
	pk1[1].p_freq = i;
 	npeak++;
    }
#endif

	/* now, working back into spect2, make a sort of "liklihood"
	 * spectrum.  Proceeding in 48ths of an octave,  from 2 to
	 * n/2 (in bins), the liklihood of each pitch range is contributed
	 * to by every peak in peaklist that's an integer multiple of it
	 * in frequency.
	 */
	 
    for (i = 0, fp1 = histogram; i < maxbin; i++) *fp1++ = 0;
    for (i = 0, pk1 = peaklist; i < npeak; i++, pk1++)
    {
	float pit = BPERO_OVER_LOG2 * flog(pk1->p_freq) - 96.0f;
	float binbandwidth = FACTORTOBINS * pk1->p_width/pk1->p_freq;
	float putbandwidth = (binbandwidth < 2 ? 2 : binbandwidth);
	float weightbandwidth = (binbandwidth < 0.5f ? 0.5f : binbandwidth);
	float weightamp = 1.0f + 3.0f * pk1->p_pow / pow;
	for (j = 0, fp2 = sigfiddle_partialonset; j < NPARTIALONSET; j++, fp2++)
	{
	    float bin = pit - *fp2;
	    if (bin < maxbin)
	    {
		float para, pphase, score = 30.0f * weightamp /
		    ((j+x->x_npartial) * weightbandwidth);
		int firstbin = bin + 0.5f - 0.5f * putbandwidth;
		int lastbin = bin + 0.5f + 0.5f * putbandwidth;
		int ibw = lastbin - firstbin;
		if (firstbin < -BINGUARD) break;
		para = 1.0f / (putbandwidth * putbandwidth);
		for (k = 0, fp3 = histogram + firstbin,
		    pphase = firstbin-bin; k <= ibw;
			k++, fp3++,  pphase += 1.0f)
		{
			    *fp3 += score * (1.0f - para * pphase * pphase);
		}
	    }
	}
    }
#if 1
    if (x->x_nprint)
    {
	for (i = 0; i < 6*5; i++)
	{
	    float fhz = hzperbin * exp ((8*i + 96) * (1./BPERO_OVER_LOG2));
	    if (!(i % 6)) post("-- bin %d pitch %f freq %f----", 8*i,
	    	ftom(fhz), fhz);;
	    post("%3d %3d %3d %3d %3d %3d %3d %3d", 
		(int)(histogram[8*i]), 
		(int)(histogram[8*i+1]), 
		(int)(histogram[8*i+2]), 
		(int)(histogram[8*i+3]), 
		(int)(histogram[8*i+4]), 
		(int)(histogram[8*i+5]), 
		(int)(histogram[8*i+6]), 
		(int)(histogram[8*i+7]));
	}
    }
    
#endif

	/*
	 * Next we find up to NHIST strongest peaks in the histogram.
	 * if a peak is related to a stronger one via an interval in
	 * the sigfiddle_partialonset array,  we suppress it.
	 */

    for (nhist = 0; nhist < x->x_nhist; nhist++)
    {
	int index;
	float best;
	if (nhist)
	{
	    for (best = 0, index = -1, j=1; j < maxbin-1; j++)
	    {
		if (histogram[j] > best && histogram[j] > histogram[j-1] &&
		    histogram[j] > histogram[j+1])
		{
		    for (k = 0; k < nhist; k++)
			if (histvec[k].h_index == j)
			    goto peaknogood;
		    for (k = 0; k < NPARTIALONSET; k++)
		    {
			if (j - sigfiddle_intpartialonset[k] < 0) break;
			if (histogram[j - sigfiddle_intpartialonset[k]]
			    > histogram[j]) goto peaknogood;
		    }
		    for (k = 0; k < NPARTIALONSET; k++)
		    {
			if (j + sigfiddle_intpartialonset[k] >= maxbin) break;
			if (histogram[j + sigfiddle_intpartialonset[k]]
			    > histogram[j]) goto peaknogood;
		    }
		    index = j;
		    best = histogram[j];
		}
	    peaknogood: ;
	    }
	}
	else
	{
	    for (best = 0, index = -1, j=0; j < maxbin; j++)
		if (histogram[j] > best)
		    index = j,  best = histogram[j];
	}
	if (index < 0) break;
	histvec[nhist].h_value = best;
	histvec[nhist].h_index = index;
    }
#if 1
    if (x->x_nprint)
    {
	for (i = 0; i < nhist; i++)
	{
	    post("index %d freq %f --> value %f", histvec[i].h_index, 
		exp((1./BPERO_OVER_LOG2) * (histvec[i].h_index + 96)), 
		histvec[i].h_value);
	    post("next %f , prev %f", 
		exp((1./BPERO_OVER_LOG2) * (histvec[i].h_index + 97)), 
		exp((1./BPERO_OVER_LOG2) * (histvec[i].h_index + 95)) );
	}
    }
#endif

	/* for each histogram peak, we now search back through the
	 * FFT peaks.  A peak is a pitch if either there are several
	 * harmonics that match it,  or else if (a) the fundamental is
	 * present,  and (b) the sum of the powers of the contributing peaks
	 * is at least 1/100 of the total power.
	 * 
	 * A peak is a contributor if its frequency is within 25 cents of
	 * a partial from 1 to 16. 
	 */

    for (i = 0; i < nhist; i++)
    {
	float cumpow = 0, cumstrength = 0, freqnum = 0, freqden = 0;
	int npartials = 0,  nbelow8 = 0;
	    /* guessed-at frequency in bins */
	float putfreq = fexp((1.0f / BPERO_OVER_LOG2) *
	    (histvec[i].h_index + 96.0f));
	for (j = 0; j < npeak; j++)
	{
	    float fpnum = peaklist[j].p_freq/putfreq;
	    int pnum = fpnum + 0.5f;
	    float fipnum = pnum;
	    float deviation;
	    if (pnum > 16 || pnum < 1) continue;
	    deviation = 1.0f - fpnum/fipnum;
	    if (deviation > -PARTIALDEVIANCE && deviation < PARTIALDEVIANCE)
	    {
		/*
		 * we figure this is a partial since it's within 1/4 of
		 * a halftone of a multiple of the putative frequency.
		 */

		float stdev, weight;
		npartials++;
		if (pnum < 8) nbelow8++;
		cumpow += peaklist[j].p_pow;
		cumstrength += sqrt(sqrt(peaklist[j].p_pow));
		stdev = (peaklist[j].p_width > MINBW ?
		    peaklist[j].p_width : MINBW);
		weight = 1.0f / ((stdev*fipnum) * (stdev*fipnum));
		freqden += weight;
		freqnum += weight * peaklist[j].p_freq/fipnum;		
#if 1
		if (x->x_nprint)
		{
		    post("peak %d partial %d f=%f w=%f", 
			j, pnum, peaklist[j].p_freq/fipnum, weight);
		}
#endif
	    }
#if 1
	    else if (x->x_nprint) post("peak %d partial %d dev %f", 
			j, pnum, deviation);
#endif
	}
	if ((nbelow8 < 4 || npartials < 7) && cumpow < 0.01f * pow)
	    histvec[i].h_value = 0;
	else
	{
	    float cumpow = (cumstrength * cumstrength) *
		(cumstrength * cumstrength);

	        /* we passed the test... save the values we got */
	    histvec[i].h_pitch = ftom(hzperbin * freqnum/freqden);
	    histvec[i].h_loud = (100.0f -DBFUDGE) + (LOGTODB) * log(cumpow/n);
	}
    }
#if 1
    if (x->x_nprint)
    {
	for (i = 0; i < nhist; i++)
	{
	    if (histvec[i].h_value)
		post("index %d pit %f loud %f", histvec[i].h_index, 
		histvec[i].h_pitch, histvec[i].h_loud);
	    else post("-- cancelled --");
	}
    }
#endif

	/* now try to find continuous pitch tracks that match the new
	 * pitches.  First mark each peak unmatched. 
	 */
    for (i = 0, hp1 = histvec; i < nhist; i++, hp1++)
	hp1->h_used = 0;

	/* for each old pitch, try to match a new one to it. */
    for (i = 0, phist = x->x_hist; i < x->x_npitch; i++,  phist++)
    {
	float thispitch = phist->h_pitches[oldphase];
	phist->h_pitch = 0;	    /* no output, thanks */
	phist->h_wherefrom = 0;
	if (thispitch == 0.0f) continue;
	for (j = 0, hp1 = histvec; j < nhist; j++, hp1++)
	    if (hp1->h_value && hp1->h_pitch > thispitch - GLISS
		&& hp1->h_pitch < thispitch + GLISS)
	{
	    phist->h_wherefrom = hp1;
	    hp1->h_used = 1;
	}
    }
    for (i = 0, hp1 = histvec; i < nhist; i++, hp1++) 
	if (hp1->h_value && !hp1->h_used)
    {
	for (j = 0, phist = x->x_hist; j < x->x_npitch; j++,  phist++)
	    if (!phist->h_wherefrom)
	{
	    phist->h_wherefrom = hp1;
	    phist->h_age = 0;
	    phist->h_noted = 0;
	    hp1->h_used = 1;
	    goto happy;
	}
	break;
    happy: ;
    }
	/* copy the pitch info into the history vector */
    for (i = 0, phist = x->x_hist; i < x->x_npitch; i++,  phist++)
    {
	if (phist->h_wherefrom)
	{
	    phist->h_amps[newphase] = phist->h_wherefrom->h_loud;
	    phist->h_pitches[newphase] =
		phist->h_wherefrom->h_pitch;
	    (phist->h_age)++;
	}
	else
	{
	    phist->h_age = 0;
	    phist->h_amps[newphase] = phist->h_pitches[newphase] = 0;
	}
    }
#if 1
    if (x->x_nprint)
    {
	post("vibrato %d %f", x->x_vibbins, x->x_vibdepth);
	for (i = 0, phist = x->x_hist; i < x->x_npitch; i++,  phist++)
	{
	    post("noted %f, age %d", phist->h_noted,  phist->h_age);
#ifndef I860
	    post("values %f %f %f %f %f",
		phist->h_pitches[newphase], 
		phist->h_pitches[(newphase + HISTORY-1)%HISTORY], 
		phist->h_pitches[(newphase + HISTORY-2)%HISTORY], 
		phist->h_pitches[(newphase + HISTORY-3)%HISTORY], 
		phist->h_pitches[(newphase + HISTORY-4)%HISTORY]);
#endif
	}
    }
#endif
	/* look for envelope attacks */

    x->x_attackvalue = 0;

    if (x->x_peaked)
    {
	if (db > x->x_amphi)
	{
	    int binlook = newphase - x->x_attackbins;
	    if (binlook < 0) binlook += HISTORY;
	    if (db > x->x_dbs[binlook] + x->x_attackthresh)
	    {
		x->x_attackvalue = 1;
		x->x_peaked = 0;
	    }
	}
    }
    else
    {
	int binlook = newphase - x->x_attackbins;
	if (binlook < 0) binlook += HISTORY;
	if (x->x_dbs[binlook] > x->x_amphi && x->x_dbs[binlook] > db)
	    x->x_peaked = 1;
    }
    
	/* for each current frequency track, test for a new note using a
	 * stability criterion.  Later perhaps we should also do as in
	 * pitch~ and check for unstable notes a posteriori when
	 * there's a new attack with no note found since the last onset;
	 * but what's an attack &/or onset when we're polyphonic?
	 */

    for (i = 0, phist = x->x_hist; i < x->x_npitch; i++,  phist++)
    {
	    /*
	     * if we've found a pitch but we've now strayed from it turn
	     * it off.
	     */
	if (phist->h_noted)
	{
	    if (phist->h_pitches[newphase] > phist->h_noted + x->x_vibdepth
		|| phist->h_pitches[newphase] < phist->h_noted - x->x_vibdepth)
		    phist->h_noted = 0;
	}
	else
	{
	    
	    if (phist->h_wherefrom && phist->h_age >= x->x_vibbins)
	    {
		float centroid = 0;
		int not = 0;
		for (j = 0, k = newphase; j < x->x_vibbins; j++)
		{
		    centroid += phist->h_pitches[k];
		    k--;
		    if (k < 0) k = HISTORY-1;
		}
		centroid /= x->x_vibbins;
		for (j = 0, k = newphase; j < x->x_vibbins; j++)
		{
			/* calculate deviation from norm */
		    float dev = centroid - phist->h_pitches[k];
		    k--;
		    if (k < 0) k = HISTORY-1;
		    if (dev > x->x_vibdepth ||
			-dev > x->x_vibdepth) not = 1;
		}
		if (!not)
		{
		    phist->h_pitch = phist->h_noted = centroid;
		}
	    }
	}
    }
    return;

nopow:
    for (i = 0; i < x->x_npitch; i++)
    {
	x->x_hist[i].h_pitch = x->x_hist[i].h_noted = 
	    x->x_hist[i].h_pitches[newphase] =
	    x->x_hist[i].h_amps[newphase] = 0;
	x->x_hist[i].h_age = 0;
    }
    x->x_peaked = 1;
    x->x_dbage = 0;
}

void sigfiddle_debug(t_sigfiddle *x)
{
    x->x_nprint = 1;
}

void sigfiddle_print(t_sigfiddle *x)
{
    post("amp-range %f %f,",  x->x_amplo, x->x_amphi);
    post("reattack %f %f,",  x->x_attacktime, x->x_attackthresh);
    post("vibrato %f %f",  x->x_vibtime, x->x_vibdepth);
    post("npartial %f",  x->x_npartial);
}

void sigfiddle_amprange(t_sigfiddle *x, double amplo,  double amphi)
{
    if (amplo < 0) amplo = 0;
    if (amphi < amplo) amphi = amplo + 1;
    x->x_amplo = amplo;
    x->x_amphi = amphi;
}

void sigfiddle_reattack(t_sigfiddle *x, double attacktime, double attackthresh)
{
    if (attacktime < 0) attacktime = 0;
    if (attackthresh <= 0) attackthresh = 1000;
    x->x_attacktime = attacktime;
    x->x_attackthresh = attackthresh;
    x->x_attackbins = (x->x_sr * 0.001 * attacktime) / x->x_hop;
    if (x->x_attackbins >= HISTORY) x->x_attackbins = HISTORY - 1;
}

void sigfiddle_vibrato(t_sigfiddle *x, double vibtime, double vibdepth)
{
    if (vibtime < 0) vibtime = 0;
    if (vibdepth <= 0) vibdepth = 1000;
    x->x_vibtime = vibtime;
    x->x_vibdepth = vibdepth;
    x->x_vibbins = (x->x_sr * 0.001  * vibtime) / x->x_hop;
    if (x->x_vibbins >= HISTORY) x->x_vibbins = HISTORY - 1;
    if (x->x_vibbins < 1) x->x_vibbins = 1;
}

void sigfiddle_npartial(t_sigfiddle *x, double npartial)
{
    if (npartial < 0.1) npartial = 0.1;
    x->x_npartial = npartial;
}

int sigfiddle_doinit(t_sigfiddle *x, long npoints, long npitch)
{
    float *buf1, *buf2,  *buf3;
    int i, hop;

    if (npoints < MINPOINTS || npoints > MAXPOINTS) npoints = DEFAULTPOINTS;
    npoints = 1 << sigfiddle_ilog2(npoints);
    hop = npoints>>1;
    if (!(buf1 = getbytes(sizeof(float) * hop)))
    {
	error("fiddle~: out of memory");
	return (0);
    }
    if (!(buf2 = getbytes(sizeof(float) * (npoints + 4 * FILTSIZE))))
    {
	freebytes(buf1, sizeof(float) * hop);
	error("fiddle~: out of memory");
	return (0);
    }
    if (!(buf3 = getbytes(sizeof(float) * npoints)))
    {
	freebytes(buf1, sizeof(float) * hop);
	freebytes(buf2, sizeof(float) * (npoints + 4 * FILTSIZE));
	error("fiddle~: out of memory");
	return (0);
    }
    for (i = 0; i < hop; i++) buf1[i] = 0;
    for (i = 0; i < npoints + 4 * FILTSIZE; i++) buf2[i] = 0;
    for (i = 0; i < hop; i++)
	buf3[2*i] =    cos((3.14159*i)/(npoints)), 
	buf3[2*i+1] = -sin((3.14159*i)/(npoints));
    if (npitch <= 0) npitch = 1;
    else if (npitch > MAXNPITCH) npitch = MAXNPITCH;

    x->x_inbuf = buf1;
    x->x_lastanalysis = buf2;
    x->x_spiral = buf3;
    x->x_phase = 0;
    x->x_histphase = 0;
    x->x_hop = npoints>>1;
    x->x_sr = 44100;		/* this and the next are filled in later */
    for (i = 0; i < MAXNPITCH; i++)
    {
	int j;
	x->x_hist[i].h_pitch = x->x_hist[i].h_noted = 0;
	x->x_hist[i].h_age = 0;
	x->x_hist[i].h_wherefrom = 0;
	x->x_hist[i].h_outlet = 0;
	for (j = 0; j < HISTORY; j++)
	    x->x_hist[i].h_amps[j] = x->x_hist[i].h_pitches[j] = 0;
    }
    x->x_nprint = 0;
    x->x_npitch = npitch;
    x->x_nhist = npitch;
    for (i = 0; i < HISTORY; i++) x->x_dbs[i] = 0;
    x->x_dbage = 0;
    x->x_peaked = 0;
    x->x_amplo = DEFAMPLO;
    x->x_amphi = DEFAMPHI;
    x->x_attacktime = DEFATTACKTIME;
    x->x_attackbins = 1;		/* real value calculated afterward */
    x->x_attackthresh = DEFATTACKTHRESH;
    x->x_vibtime = DEFVIBTIME;
    x->x_vibbins = 1;			/* real value calculated afterward */
    x->x_vibdepth = DEFVIBDEPTH;
    x->x_npartial = 7;
    x->x_attackvalue = 0;
    return (1);
}

/* formalities */
static t_int *fiddle_perform(t_int *w)
{
    t_float *in = (t_float *)(w[1]);
    t_sigfiddle *x = (t_sigfiddle *)(w[2]);
    int n = (int)(w[3]);
    int count;
    float *fp;
    for (count = 0, fp = x->x_inbuf + x->x_phase; count < n; count++)
    	*fp++ = *in++;
    if (fp == x->x_inbuf + x->x_hop)
    {
	sigfiddle_doit(x);
	x->x_phase = 0;
	clock_delay(x->x_clock, 0L);
	if (x->x_nprint) x->x_nprint--;
    }
    else x->x_phase += n;
    return (w+4);
}

void sigfiddle_dsp(t_sigfiddle *x, t_signal **sp)
{
    x->x_sr = sp[0]->s_sr;
    sigfiddle_reattack(x, x->x_attacktime, x->x_attackthresh);
    sigfiddle_vibrato(x, x->x_vibtime, x->x_vibdepth);
    dsp_add(fiddle_perform, 3, sp[0]->s_vec, x, sp[0]->s_n);
}

void sigfiddle_tick(t_sigfiddle *x)	/* callback function for the clock */
{
    int i;
    t_pitchhist *ph;
    outlet_float(x->x_envout, x->x_dbs[x->x_histphase]);
    for (i = 0,  ph = x->x_hist; i < x->x_npitch; i++,  ph++)
    {
	t_atom at[3];
	SETFLOAT(at, ph->h_pitches[x->x_histphase]);
	SETFLOAT(at+1, ph->h_amps[x->x_histphase]);
	outlet_list(ph->h_outlet, gensym("list"), 2, at);
    }
    if (x->x_attackvalue) outlet_bang(x->x_attackout);
    for (i = 0,  ph = x->x_hist; i < x->x_npitch; i++,  ph++)
 	if (ph->h_pitch) outlet_float(x->x_noteout, ph->h_pitch);
}

void sigfiddle_ff(t_sigfiddle *x)		/* cleanup on free */
{
    if (x->x_inbuf)
    {
    	freebytes(x->x_inbuf, sizeof(float) * x->x_hop);
    	freebytes(x->x_lastanalysis, sizeof(float) * (2*x->x_hop + 4 * FILTSIZE));
    	freebytes(x->x_spiral, sizeof(float) * 2*x->x_hop);
    	clock_free(x->x_clock);
    }
}

static t_class *sigfiddle_class;

void *sigfiddle_new(float npoints, float npitch)
{
    t_sigfiddle *x = (t_sigfiddle *)pd_new(sigfiddle_class);
    int i;
    
    if (!sigfiddle_doinit(x, (long) npoints, (long) npitch))
    {
    	x->x_inbuf = 0;     /* prevent the free routine from cleaning up */
    	pd_free(&x->x_ob.ob_pd);
    	return (0);
    }
    x->x_envout = outlet_new(&x->x_ob, gensym("float"));
    for (i = 0; i < npitch; i++)
	 x->x_hist[i].h_outlet = outlet_new(&x->x_ob, gensym("list"));
    x->x_attackout = outlet_new(&x->x_ob, gensym("bang"));
    x->x_noteout = outlet_new(&x->x_ob, gensym("float"));
    x->x_clock = clock_new(&x->x_ob.ob_pd, sigfiddle_tick);
    return (x);
}

void fiddle_setup()
{
  sigfiddle_class = class_new(gensym("fiddle"), sigfiddle_new, sigfiddle_ff,
					          sizeof(t_sigfiddle), 0, A_DEFFLOAT, A_DEFFLOAT, 0);
    class_addmethod(sigfiddle_class, sigfiddle_dsp, gensym("dsp"), 0);
    class_addmethod(sigfiddle_class, sigfiddle_debug, 
						  gensym("debug"), A_DEFFLOAT, 0);
    class_addmethod(sigfiddle_class, sigfiddle_amprange, 
						  gensym("amp-range"), A_FLOAT, A_FLOAT, 0);
    class_addmethod(sigfiddle_class, sigfiddle_reattack, 
						  gensym("reattack"), A_FLOAT, A_FLOAT, 0);
    class_addmethod(sigfiddle_class, sigfiddle_vibrato, 
						  gensym("vibrato"), A_FLOAT, A_FLOAT, 0);
    class_addmethod(sigfiddle_class, sigfiddle_npartial, 
						  gensym("npartial"), A_FLOAT, 0);
    class_addmethod(sigfiddle_class, sigfiddle_print, gensym("print"), 0);
    class_addmethod(sigfiddle_class, nullfn, gensym("signal"), 0);
    {
      int i;
      for (i = 0; i < MAXPOSSIBLEBIN; i++)
	sigfiddle_binbin[i] = 4 + 48 * log(i)/log(2.);
    }
}



